from iolanta.facet import Facet
from octadocs.iolanta import render
from octadocs_table.models import TABLE


class Th(Facet):
    """Render a table column header."""

    def render(self):
        """Render the column."""
        return render(
            node=self.iri,
            octiron=self.octiron,
            environments=[TABLE.th],
        )



from typing import TextIO

import yaml
from iolanta.convert_dollar_signs import convert_dollar_signs
from iolanta.models import LDDocument
from iolanta.parsers.base import Parser

try:  # noqa
    from yaml import CSafeLoader as SafeLoader  # noqa
except ImportError:
    from yaml import SafeLoader  # type: ignore   # noqa


class YAML(Parser):
    """Load YAML data."""

    def as_jsonld_document(self, raw_data: TextIO) -> LDDocument:
        """Read YAML content and adapt it to JSON-LD format."""
        document = yaml.load(raw_data, Loader=SafeLoader)
        return convert_dollar_signs(document)



from pathlib import Path
from typing import Iterable, Type

from iolanta.loaders.local_file import Loader, LocalFile
from iolanta.models import LDContext, LDDocument, Quad
from rdflib import URIRef
from urlpath import URL


def choose_loader_by_url(url: URL) -> Type[Loader]:
    """Find loader by URL scheme."""
    return LocalFile


def as_document(url: URL) -> LDDocument:
    """Retrieve the document presented by the specified URL."""
    loader_class = choose_loader_by_url(url)
    return loader_class().as_jsonld_document(url)


def as_quad_stream(
    url: URL,
    iri: URIRef,
    default_context: LDContext,
    root_directory: Path,
) -> Iterable[Quad]:
    """Retrieve the stream presented by the specified URL."""
    loader_class = choose_loader_by_url(url)
    return loader_class(
        root_directory=root_directory,
        default_context=default_context,
    ).as_quad_stream(
        url=url,
        iri=iri,
    )



from pathlib import Path
from typing import Any, Dict, Iterator, Optional

import rdflib
from octadocs.types import Context, Triple


class Loader:
    """Data importer for Octiron."""

    # Which files is this loader working with?
    regex: str

    # Absolute path to source file
    path: Path

    # Local address of the file, which will be used as graph name
    local_iri: rdflib.URIRef

    # The URL of the page (relative or absolute) under which the page will be
    # accessible for users.
    global_url: Optional[str]

    # JSON-LD context
    context: Context

    # Named contexts from Octiron config
    named_contexts: Optional[Dict[str, Any]] = None

    def __init__(
        self,
        path: Path,
        local_iri: rdflib.URIRef,
        global_url: Optional[str],
        context: Context,
        named_contexts: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Initialize the data loader."""
        self.path = path
        self.context = context
        self.local_iri = local_iri
        self.global_url = global_url
        self.named_contexts = named_contexts

    def stream(self) -> Iterator[Triple]:
        """Read the source data and return a stream of triples."""
        raise NotImplementedError()



from dataclasses import dataclass
from functools import cached_property
from typing import Optional

from octadocs.octiron import Octiron
from rdflib.term import Node, URIRef


@dataclass
class Facet:
    """Base facet class."""

    iri: Node
    octiron: Octiron
    environment: Optional[URIRef] = None

    @cached_property
    def uriref(self) -> URIRef:
        """Format as URIRef."""
        return URIRef(self.iri)

    def query(self, query_text: str, **kwargs):
        """SPARQL query."""
        return self.octiron.query(
            query_text=query_text,
            **kwargs
        )

    def render(self):
        """Render the facet."""
        raise NotImplementedError()

    def __str__(self):
        """Render."""
        return str(self.render())



import operator
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Tuple

from iolanta import as_document
from iolanta.models import LDContext, Triple
from mkdocs.plugins import BasePlugin
from octadocs.mixins import OctadocsMixin
from octadocs_github.cli import app, logger
from rdflib import RDFS, Literal, Namespace, URIRef
from typer import Typer
from urlpath import URL

GH = Namespace('https://octadocs.io/github/')


@dataclass
class Repo:
    """Repository description."""

    url: str

    @property
    def url_parts(self) -> Tuple[str, str]:
        """Parse the GitHub URL."""
        try:
            _hostname, owner_name, repo_name, *etc = URL(self.url).parts
        except ValueError as err:
            raise ValueError(
                f'{self.url} is not a valid GitHub repo URL.',
            ) from err

        return owner_name, repo_name

    @property
    def owner_name(self):
        return self.url_parts[0]

    @property
    def repo_name(self):
        return self.url_parts[1]


class GithubPlugin(OctadocsMixin, BasePlugin):
    """Render an HTML table from data presented in the graph."""

    namespaces = {'gh': GH}
    plugin_data_dir = Path(__file__).parent / 'data'

    def typer(self) -> Typer:
        """Typer instance."""
        return app

    def named_contexts(self) -> Dict[str, LDContext]:
        """Reusable named contexts."""
        return {
            'github': as_document(
                URL('file://', self.plugin_data_dir / 'named-context.yaml'),
            ),
        }

    def vocabularies(self) -> Dict[URIRef, Path]:
        """Load Github inference rules."""
        return {
            URIRef(GH): self.plugin_data_dir / 'github.yaml',
        }

    def inference(self):
        """Find all GitHub URLs and mark them."""
        logger.warning('Looking for GitHub URLs in the graph...')
        rows = self.query(
            '''
            SELECT DISTINCT ?url WHERE {
                {
                    ?url ?p ?o
                } UNION {
                    ?s ?p ?url
                }

                FILTER isIRI(?url) .

                FILTER(
                    STRSTARTS(
                        str(?url),
                        "https://github.com/"
                    )
                ) .
            }
            ''',
        )

        iris = map(operator.itemgetter('url'), rows)

        for iri in iris:
            url = URL(iri)

            try:
                _hostname, owner_name, repo_name, *etc = url.parts
            except ValueError:
                continue

            if etc:
                # This is not a repo name.
                continue

            owner_iri = URIRef(f'https://github.com/{owner_name}')
            repo_label = f'{owner_name}/{repo_name}'

            self.insert(
                Triple(iri, RDFS.label, Literal(repo_label)),
                Triple(iri, GH.owner, owner_iri),
                Triple(owner_iri, RDFS.label, Literal(owner_name)),
            )



import json
import logging
from pathlib import Path

from github import Github, UnknownObjectException
from ldflex import LDFlex
from octadocs.storage import load_graph
from octadocs_github.models import RepoDescription
from rdflib import XSD
from typer import Typer
from urlpath import URL

logger = logging.getLogger(__name__)
app = Typer(
    name='github',
    help='Manage data from GitHub.',
)


def ldflex_from_cache() -> LDFlex:
    graph = load_graph(Path.cwd() / '.cache/octadocs')
    return LDFlex(graph=graph)


def describe_repo(
    name: str,
    github_api_base_url: URL,
) -> RepoDescription:
    """
    Describe a repo by its full name.

    Full name format: {owner}/{repo}.
    """
    github = Github()
    repo = github.get_repo(name)
    raise ValueError(repo)


@app.command(name='update')
def github_cli():
    """Update GitHub information."""
    ldflex = ldflex_from_cache()
    rows = ldflex.query(
        '''
        SELECT ?repo_name WHERE {
            ?repo
                a gh:Repo ;
                rdfs:label ?repo_name .
        }
        '''
    )

    gh = Github()

    docs_dir = Path.cwd() / 'docs'

    if not docs_dir.is_dir():
        raise ValueError(
            f'{docs_dir} is considered to be docs directory but it does not '
            f'exist.',
        )

    target_dir = docs_dir / 'generated/octadocs-github'
    target_dir.mkdir(parents=True, exist_ok=True)
    (target_dir / 'context.json').write_text(json.dumps(
        {
            'updated_at': {
                '@type': 'xsd:dateTime',
            },
            'pushed_at': {
                '@type': 'xsd:dateTime',
            },
            '@vocab': 'https://octadocs.io/github/',
            'license': {
                '@type': 'gh:License',
            }
            # '@import': 'github',
        },
        indent=2,
    ))

    for row in rows:
        try:
            repo = gh.get_repo(row['repo_name'])
        except UnknownObjectException:
            logger.error(
                '%s is not a valid GitHub repository name.',
                row['repo_name'],
            )
            continue

        formatted_data = repo.raw_data
        formatted_data['@id'] = repo.html_url

        file_name = repo.full_name.replace('/', '__')
        (target_dir / f'{file_name}.json').write_text(json.dumps(
            formatted_data,
            indent=2,
        ))



from dataclasses import dataclass

from documented import DocumentedError
from iolanta.facet import Facet
from more_itertools import first
from octadocs.iolanta import render
from octadocs_table.models import TABLE
from rdflib import URIRef


@dataclass
class MissingTableInstanceClass(DocumentedError):
    """
    Table does not have an instance class defined.

    Table IRI: `{self.table_iri}`

    For this table, you had tried to render `table:self` column. This cannot be
    executed at this time because the table does not have an associated
    `table:class`. Please create one:

    ```yaml
    $context:
      $import: table
    $id: {self.table_iri}
    class: <class of things you wish to display in your table>
    ```
    """

    table_iri: URIRef


class SelfFacet(Facet):
    """Render table:self property."""

    def render(self):
        """Call rendering for the underlying class of the table."""
        if self.environment is None:
            raise ValueError(
                f'Facet {self} was called with no environment specified.',
            )

        rows = self.query(
            query_text='''
            SELECT * WHERE {
                $table table:class ?cls .
            }
            ''',
            table=self.environment,
        )

        try:
            instance_class = first(rows)['cls']
        except ValueError as err:
            raise MissingTableInstanceClass(
                table_iri=self.environment,
            ) from err

        return render(
            node=instance_class,
            octiron=self.octiron,
            environments=[TABLE.th],
        )



import json
from functools import partial
from itertools import starmap
from typing import Iterable, Optional, TextIO

from iolanta.models import LDContext, LDDocument, Quad, Triple
from iolanta.parsers.base import Parser
from iolanta.reformat_blank_nodes import reformat_blank_nodes
from rdflib import Graph, URIRef


class JSON(Parser):
    """Load JSON data."""

    def as_jsonld_document(self, raw_data: TextIO) -> LDDocument:
        """Read JSON content as a JSON-LD document."""
        return json.load(raw_data)

    def as_quad_stream(
        self,
        raw_data: TextIO,
        iri: Optional[URIRef],
        context: LDContext,
    ) -> Iterable[Quad]:
        """Read JSON-LD data into a quad stream."""
        graph = Graph()
        graph.parse(
            data=raw_data.read(),
            format='json-ld',
            context=context,
        )

        triples = map(
            partial(
                reformat_blank_nodes,
                f'{iri}/{self.blank_node_prefix}',
            ),
            starmap(
                Triple,
                iter(graph),
            ),
        )

        return list(
            triple.as_quad(iri)
            for triple in triples
        )



from dominate.tags import a, code
from more_itertools import first
from octadocs.octiron import Octiron
from octadocs_adr.models import ADRNumberInvalid
from rdflib import URIRef


def link_to_adr(octiron: Octiron, iri: URIRef) -> a:
    """Link to an ADR document given by the IRI."""
    descriptions = octiron.query(
        '''
        SELECT * WHERE {
            ?page
                octa:url ?url ;
                octa:title ?label ;
                adr:number ?number .

            OPTIONAL {
                ?page adr:status / octa:symbol ?symbol .
            }
        } ORDER BY ?number LIMIT 1
        ''',
        page=iri,
    )
    location = first(descriptions, None)

    if not location:
        raise ValueError(f'Page not found by IRI: {iri}')

    number = location['number'].value

    if not isinstance(number, int):
        raise ADRNumberInvalid(
            number=number,
            page=iri,
        )

    readable_number = f'ADR{number:03}'

    if symbol := location.get('symbol'):
        readable_number = f'{symbol} {readable_number}'

    return a(
        code(readable_number),
        ' ',
        location['label'],
        href=location['url'],
    )



from dominate.tags import a  # noqa: WPS347
from iolanta.facet import Facet
from more_itertools import first
from rdflib.term import Literal


class Default(Facet):
    """Default facet for rendering an unknown node."""

    def render(self):   # noqa: C901
        """Render HTML."""
        if isinstance(self.iri, Literal):
            return str(self.iri.value)

        descriptions = self.query(
            query_text='''
            SELECT * WHERE {
                OPTIONAL {
                    ?page rdfs:label ?label .
                }

                OPTIONAL {
                    ?page octa:symbol ?symbol .
                }

                OPTIONAL {
                    ?page octa:url ?url .
                }

                OPTIONAL {
                    ?page a octa:Page .
                    BIND(true AS ?is_page)
                }
            } ORDER BY ?label LIMIT 1
            ''',
            page=self.iri,
        )

        try:
            description = first(descriptions)
        except ValueError:
            return str(self.iri)

        label = description.get('label', str(self.iri))
        url = description.get('url')

        symbol = description.get('symbol')
        if not symbol:
            symbol = ''

        if url:
            return a(
                symbol,
                label,
                href=url,
            )

        return label



from pathlib import Path
from typing import Iterator

from octadocs.types import LOCAL as local
from octadocs.types import OCTA as octa
from octadocs.types import Triple
from rdflib import RDF as rdf
from rdflib import Literal, URIRef


def describe_documentation_page(
    iri: URIRef,
    path: Path,
    docs_dir: Path,
) -> Iterator[Triple]:
    """Describe the file properties and hierarchy for the given page."""
    yield Triple(iri, octa.fileName, Literal(path.name))

    try:
        relative_path = path.relative_to(docs_dir)
    except ValueError:
        return

    try:
        breadcrumbs = list(relative_path.parents)[:-1]
    except ValueError:
        return

    children = [relative_path, *breadcrumbs]
    parents = [*children[1:], None]

    child_per_parent = zip(children, parents)

    for child, parent in child_per_parent:
        child_iri = local.term(str(child or ''))
        parent_iri = local.term(str(parent or ''))

        yield Triple(parent_iri, rdf.type, octa.Directory)
        yield Triple(child_iri, octa.isChildOf, parent_iri)

        if parent:
            yield Triple(parent_iri, octa.fileName, Literal(parent.name))




from enum import Enum
from logging import config as logging_config
from types import MappingProxyType
from typing import Any, Dict, List, NamedTuple, Optional, Union

import rdflib
from ldflex.ldflex import SelectResult
from rdflib import Graph, term
from typing_extensions import Protocol

OCTA = rdflib.Namespace('https://octadocs.io/')
IOLANTA = rdflib.Namespace('https://iolanta.tech/')
LOCAL = rdflib.Namespace('local:')


# To prevent WPS401 Found mutable module constant
DEFAULT_NAMESPACES = MappingProxyType({
    'octa': OCTA,
    'iolanta': IOLANTA,
    '': LOCAL,

    'rdf': rdflib.RDF,
    'rdfs': rdflib.RDFS,
    'xsd': rdflib.XSD,
    'schema': rdflib.SDO,
    'sh': rdflib.SH,
    'skos': rdflib.SKOS,
    'owl': rdflib.OWL,
    'dc': rdflib.DC,
    'dcat': rdflib.DCAT,
    'dcterms': rdflib.DCTERMS,
    'foaf': rdflib.FOAF,

    'table': rdflib.Namespace('https://octadocs.io/table/'),
    'prov': rdflib.URIRef(rdflib.PROV),
})


DEFAULT_CONTEXT = MappingProxyType({
    '@vocab': LOCAL,
    '@base': LOCAL,

    'label': 'rdfs:label',
    'comment': 'rdfs:comment',
    'rdfs:isDefinedBy': {
        '@type': '@id',
    },
    'rdfs:subClassOf': {
        '@type': '@id',
    },
    'octa:subjectOf': {
        '@type': '@id',
    },

    'title': 'octa:title',
    'position': 'octa:position',

    # The default namespaces list should be included in context
    # We have to convert URLs to strings though - to make them serializable.
    **{
        namespace_name: str(namespace_url)
        for namespace_name, namespace_url
        in DEFAULT_NAMESPACES.items()
        if namespace_name
    },
})


class Triple(NamedTuple):
    """RDF triple."""

    subject: rdflib.URIRef
    predicate: rdflib.URIRef
    object: Union[rdflib.URIRef, rdflib.Literal]  # noqa: WPS125

    def as_quad(self, graph: rdflib.URIRef) -> 'Quad':
        """Add graph to this triple and hence get a quad."""
        return Quad(self.subject, self.predicate, self.object, graph)


class Quad(NamedTuple):
    """Triple assigned to a named graph."""

    subject: rdflib.URIRef
    predicate: rdflib.URIRef
    object: Union[rdflib.URIRef, rdflib.Literal]  # noqa: WPS125
    graph: rdflib.URIRef


# Should be Dict[str, 'Context'] but mypy crashes on a recursive type.
Context = Optional[   # type: ignore
    Union[str, int, float, Dict[str, Any]]
]


logging_config.dictConfig({
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s',
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'formatter': 'standard',
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': False,
        },
    },
})
SelectRow = Dict[str, term.Node]


QueryResult = Union[
    SelectResult,   # SELECT
    Graph,          # CONSTRUCT
    bool,           # ASK
]


class Query(Protocol):
    """Query protocol."""

    def __call__(
        self,
        query_text: str,
        instance: rdflib.ConjunctiveGraph,
        **kwargs: str,
    ) -> QueryResult:
        """Query prototype."""


class QueryResultsFormat(str, Enum):
    """Format to print query results in CLI."""

    PRETTY = 'pretty'
    CSV = 'csv'
    JSON = 'json'



from dataclasses import dataclass
from typing import Dict, List, Optional, Union

from rdflib import Graph
from rdflib.plugins.sparql.processor import SPARQLResult
from rdflib.term import Identifier, Node, Variable

SelectRow = Dict[str, Node]


class SelectResult(List[SelectRow]):
    """Result of a SPARQL SELECT."""

    @property
    def first(self) -> Optional[SelectRow]:
        """Return first element of the list."""
        return self[0] if self else None


SPARQLQueryArgument = Optional[Union[Node, str, int, float]]


QueryResult = Union[
    SelectResult,   # SELECT
    Graph,          # CONSTRUCT
    bool,           # ASK
]


def _format_query_bindings(
    bindings: List[Dict[Variable, Identifier]],
) -> SelectResult:
    """
    Format bindings before returning them.

    Converts Variable to str for ease of addressing.
    """
    return SelectResult([
        {
            str(variable_name): rdf_value
            for variable_name, rdf_value
            in row.items()
        }
        for row in bindings
    ])


@dataclass
class LDFlex:
    """Fluent interface to a semantic graph."""

    graph: Graph

    def query(
        self,
        query_text: str,
        **kwargs: SPARQLQueryArgument,
    ) -> QueryResult:
        """
        Run a SPARQL `SELECT`, `CONSTRUCT`, or `ASK` query.

        Args:
            query_text: The SPARQL text;
            **kwargs: bind variables in the query to values if necessary. For
                example:

                ```python
                ldflex.query(
                    'SELECT ?title WHERE { ?page octa:title ?title }',
                    ?page=page_iri,
                )
                ```

        Returns:
            Results of the query:

            - a graph for `CONSTRUCT`,
            - a list of dicts for `SELECT`,
            - or a boolean for `ASK`.
        """
        sparql_result: SPARQLResult = self.graph.query(
            query_text,
            initBindings=kwargs,
        )

        if sparql_result.askAnswer is not None:
            return sparql_result.askAnswer

        if sparql_result.graph is not None:
            graph: Graph = sparql_result.graph
            for prefix, namespace in self.graph.namespaces():
                graph.bind(prefix, namespace)

            return graph

        return _format_query_bindings(sparql_result.bindings)



from abc import ABC
from dataclasses import dataclass
from typing import Iterable, Optional, TextIO, Type

from iolanta.models import LDContext, LDDocument, Quad
from iolanta.namespaces import PYTHON
from iolanta.parsers.base import Parser
from rdflib import URIRef
from urlpath import URL


def term_for_python_class(cls: type) -> URIRef:
    """Construct term for Python class."""
    return PYTHON.term(f'{cls.__module__}.{cls.__qualname__}')


@dataclass(frozen=True)
class Loader(ABC):
    """
    Base class for loaders.

    Loader receives a URL (or a path) to certain location. It is responsible for
    reading data from that location and returning it as a stream of RDF quads.

    Usually, depending on the data format, Loader leverages Parsers for that
    purpose.
    """

    @classmethod
    def loader_class_iri(cls) -> URIRef:
        """Import path to the loader class."""
        return term_for_python_class(cls)

    def choose_parser_class(self, url: URL) -> Type[Parser]:
        """Find which parser class to use for this URL."""
        raise NotImplementedError()

    def as_jsonld_document(
        self,
        url: URL,
        iri: Optional[URIRef] = None,
    ) -> LDDocument:
        """Represent a file as a JSON-LD document."""
        raise NotImplementedError()

    def as_file(self, url: URL) -> TextIO:
        """Construct a file-like object."""
        raise NotImplementedError()

    def as_quad_stream(
        self,
        url: str,
        iri: Optional[URIRef],
    ) -> Iterable[Quad]:
        """Convert data into a stream of RDF quads."""
        raise NotImplementedError()

    def find_context(self, url: str) -> LDContext:
        """Find context for the file."""
        raise NotImplementedError()

    def __call__(self, url: str):
        raise NotImplementedError('This is for compatibility with PYLD.')



import sys
from pathlib import Path
from typing import Optional

from ldflex import LDFlex
from octadocs.cli.formatters.csv import csv_print
from octadocs.cli.formatters.json import print_json
from octadocs.cli.formatters.pretty import pretty_print
from octadocs.storage import load_graph
from octadocs.types import QueryResultsFormat
from typer import Argument, Option, Typer

app = Typer(name='sparql')


@app.callback(invoke_without_command=True)
def sparql(
    fmt: QueryResultsFormat = Option(
        default=QueryResultsFormat.PRETTY,
        metavar='format',
    ),
    query_text: Optional[str] = Argument(
        None,
        metavar='query',
        help='SPARQL query text. Will be read from stdin if empty.',
    ),
) -> None:
    """Run a SPARQL query against the graph."""
    if query_text is None:
        query_text = sys.stdin.read()

    graph = load_graph(Path.cwd() / '.cache/octadocs')
    ldflex = LDFlex(graph)

    query_result = ldflex.query(query_text)

    {
        QueryResultsFormat.CSV: csv_print,
        QueryResultsFormat.PRETTY: pretty_print,
        QueryResultsFormat.JSON: print_json,
    }[fmt](query_result)   # type: ignore



import datetime
from typing import Any, TypeVar

from boltons.iterutils import remap

Data = TypeVar('Data')


def _convert(term: Any) -> Any:  # type: ignore
    """Convert $statement to @statement."""
    if isinstance(term, str) and term.startswith('$'):
        return '@' + term[1:]  # noqa: WPS336

    # pyld cannot expand() a document which contains data which cannot be
    # trivially serialized to JSON.
    # FIXME we may want to replace this with an xsd:date declaration, but let's
    #   revisit that later.
    if isinstance(term, (datetime.date, datetime.datetime)):
        return str(term)

    return term


def convert_dollar_signs(
    meta_data: Data,
) -> Data:
    """
    Convert $ character to @ in keys.

    We use $ by convention to avoid writing quotes.
    """
    if not meta_data:
        return {}

    return remap(
        meta_data,
        lambda path, key, value: (  # noqa: WPS110
            _convert(key),
            _convert(value),
        ),
    )



from dominate.tags import a, img, span
from iolanta.facet import Facet
from more_itertools import first
from urlpath import URL


class GithubRepo(Facet):
    """Render a link to a GitHub repo."""

    def octicon_path(self, name: str):
        """Path to octicon."""
        return URL(
            'https://raw.githubusercontent.com/squidfunk/mkdocs-material/master'
            '/material/.icons/octicons',
        ) / f'{name}.svg'

    @property
    def icon(self):
        return span(
            img(
                src=self.octicon_path('mark-github-16'),
            ),
            cls='twemoji',
        )

    def render(self):
        """Render as a link."""
        label = first(self.query(
            '''
            SELECT ?label WHERE { ?repo rdfs:label ?label }
            ''',
            repo=self.uriref,
        ))['label']

        return a(
            self.icon,
            ' ',
            label,
            href=self.iri,
            target='_blank',
        )



import json
from dataclasses import dataclass
from functools import partial
from itertools import starmap
from typing import Any, Dict, Iterator, List, Union

import rdflib
from documented import DocumentedError
from iolanta.context import merge
from iolanta.convert_dollar_signs import convert_dollar_signs
from iolanta.models import LDContext, LDDocument
from iolanta.reformat_blank_nodes import reformat_blank_nodes
from octadocs.pyld_document_loaders.choice import ChoiceLoader
from octadocs.pyld_document_loaders.merge import MergeLoader
from octadocs.pyld_document_loaders.named import NamedLoader
from octadocs.types import LOCAL, Context, Triple
from pyld import jsonld
from pyld.documentloader.requests import requests_document_loader
from pyld.jsonld import JsonLdError

try:  # noqa
    from yaml import CSafeDumper as SafeDumper  # noqa
    from yaml import CSafeLoader as SafeLoader  # noqa
except ImportError:
    from yaml import SafeDumper  # type: ignore   # noqa
    from yaml import SafeLoader  # type: ignore   # noqa

MetaData = Union[List[Dict[str, Any]], Dict[str, Any]]  # type: ignore  # noqa


@dataclass
class ExpandError(DocumentedError):
    """
    JSON-LD expand operation failed.

    Data: {self.formatted_data}
    Context: {self.formatted_context}
    IRI: {self.iri}

    {self.message}
    """

    message: str
    document: LDDocument
    context: LDContext
    iri: str

    @property
    def formatted_data(self) -> str:
        """Format document for printing."""
        return json.dumps(self.document, indent=2)

    @property
    def formatted_context(self):
        """Format context for printing."""
        return json.dumps(self.context, indent=2)


def as_triple_stream(
    raw_data: MetaData,
    context: Context,
    local_iri: str,
    named_contexts: Dict[str, Any],
    blank_node_prefix: str = '',
) -> Iterator[Triple]:
    """Convert YAML dict to a stream of triples."""
    if isinstance(raw_data, list):
        yield from _list_as_triple_stream(
            context=context,
            local_iri=local_iri,
            raw_data=raw_data,
            named_contexts=named_contexts,
            blank_node_prefix=blank_node_prefix,
        )

    elif isinstance(raw_data, dict):
        yield from _dict_as_triple_stream(
            context=context,
            local_iri=local_iri,
            raw_data=raw_data,
            named_contexts=named_contexts,
            blank_node_prefix=blank_node_prefix,
        )

    else:
        raise ValueError(f'Format of data not recognized: {raw_data}')


def jsonld_expand(data, context, local_iri, pyld_loader=None):
    """Expand the JSON_LD data with the help of particular loader."""
    options = {
        'base': str(LOCAL),
    }

    if context:
        options.update({'expandContext': context})

    if pyld_loader is not None:
        options.update({'documentLoader': pyld_loader})

    try:
        return jsonld.expand(
            data,
            options=options,
        )
    except (TypeError, JsonLdError) as err:
        raise ExpandError(
            message=str(err),
            document=data,
            context=context,
            iri=local_iri,
        ) from err


def _dict_as_triple_stream(  # type: ignore
    raw_data: LDDocument,
    context: LDContext,
    local_iri: str,
    named_contexts: Dict[str, LDContext],
    blank_node_prefix: str = '',
) -> Iterator[Triple]:
    """Convert dict into a triple stream."""
    meta_data = convert_dollar_signs(raw_data)

    local_context = meta_data.pop('@context', None)
    if isinstance(local_context, str):
        try:
            local_context = named_contexts[local_context]
        except KeyError as err:
            raise ValueError(
                f'Named context {local_context} is unknown.',
            ) from err

    if local_context is not None:
        context = merge(
            first=context,
            second=local_context,
        )

    if meta_data.get('@id') is None:
        # The document author did not tell us about what their document is.
        # In this case, we assume that the local_iri of the document file
        # is the subject of the document description.
        meta_data['@id'] = local_iri

    # The author specified an IRI the document tells us about. Let us
    # link this IRI to the local document IRI.
    meta_data['subjectOf'] = local_iri

    # Reason: https://github.com/RDFLib/rdflib-jsonld/issues/97
    # If we don't expand with an explicit @base, import will fail silently.
    meta_data = jsonld_expand(
        data=meta_data,
        local_iri=local_iri,
        context=context,
        pyld_loader=MergeLoader(
            base_context=context,
            loader=ChoiceLoader(loaders=[
                NamedLoader(documents=named_contexts),
                requests_document_loader(),
            ]),
        )
    )

    # Reason: https://github.com/RDFLib/rdflib-jsonld/issues/98
    # If we don't flatten, @included sections will not be imported.
    meta_data = jsonld.flatten(meta_data)
    serialized_meta_data = json.dumps(meta_data, indent=4)

    graph = rdflib.Graph()
    graph.parse(
        data=serialized_meta_data,
        format='json-ld',
    )
    yield from map(
        partial(
            reformat_blank_nodes,
            f'{local_iri}/{blank_node_prefix}',
        ),
        starmap(
            Triple,
            iter(graph),
        ),
    )


def _list_as_triple_stream(  # type: ignore
    raw_data: List[Dict[str, Any]],
    context: Context,
    local_iri: str,
    named_contexts: Dict[str, Any],
    blank_node_prefix: str = '',
) -> Iterator[Triple]:
    """Convert a list into a triple stream."""
    for sub_document_number, sub_document in enumerate(raw_data):
        yield from as_triple_stream(
            raw_data=sub_document,
            context=context,
            local_iri=local_iri,
            named_contexts=named_contexts,
            blank_node_prefix=f'{blank_node_prefix}_{sub_document_number}',
        )



from functools import cached_property
from typing import Optional, cast

from iolanta.facets.base import FacetSearchAttempt
from more_itertools import first
from rdflib.term import URIRef


class FindFacetByEnvironment(FacetSearchAttempt):
    """
    Find facet by environment.

    Look for such `?facet` that `{self.environment}` `iolanta:hasDefaultFacet`
    `?facet`.
    """

    @cached_property
    def facet(self) -> Optional[URIRef]:
        """Find facet."""
        rows = self.ldflex.query(
            '''
            SELECT ?facet WHERE {
                $env iolanta:hasDefaultFacet ?facet .
            }
            ''',
            env=self.environment,
        )

        try:
            return cast(URIRef, first(rows)['facet'])
        except (ValueError, TypeError):
            return None



from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, List

import frontmatter
import rdflib
from documented import DocumentedError
from octadocs.octiron.plugins import Loader
from octadocs.octiron.yaml_extensions import as_triple_stream
from octadocs.types import OCTA, Triple
from yaml.scanner import ScannerError


@dataclass
class InvalidFrontmatter(DocumentedError):
    """
    Markdown Front Matter is not a valid YAML document.

    Path: {self.path}

    Read raw front matter below:

    {self.raw_frontmatter}

    Error:

    {self.error}

    {self.explanation}
    """

    path: Path
    raw_frontmatter: str
    error: ScannerError

    @property
    def explanation(self) -> str:
        """Explain the error to the user."""
        if 'found character \'\\t\' that' in str(self.error):   # noqa: WPS342
            return 'Suggestion:\n  Use spaces, not tabs, for indentation.'

        if self.error.problem == 'mapping values are not allowed here':
            return (
                'Suggestion:\n'
                '  If property value contains ":", use quotes.\n\n'
                '  Bad:\n\n'
                '    label: FP for Sceptics: Intuitive guide to map/flatmap\n\n'
                '  Good:\n\n'
                '    label: "FP for Sceptics: Intuitive guide to map/flatmap"'
            )

        return ''


def process_yaml_scanner_error(
    err: ScannerError,
    path: Path,
) -> InvalidFrontmatter:
    """Process YAML error and explain it to the user."""
    # Read the head of the file to print it in the error message
    raw_frontmatter: List[str] = []
    with path.open() as raw_file:
        for line in raw_file:
            if len(raw_frontmatter) > 10:
                raw_frontmatter.append('...')
                break

            if line.startswith('---') and raw_frontmatter:
                raw_frontmatter.append(line.strip('\n'))
                break

            raw_frontmatter.append(line.strip('\n'))

    return InvalidFrontmatter(
        path=path,
        raw_frontmatter='\n'.join(raw_frontmatter),
        error=err,
    )


class MarkdownLoader(Loader):
    """Load semantic data from Markdown front matter."""

    regex = r'\.md$'

    def stream(self) -> Iterator[Triple]:
        """Return stream of triples."""
        try:
            meta_data = frontmatter.load(self.path).metadata
        except ScannerError as err:
            raise process_yaml_scanner_error(
                err=err,
                path=self.path,
            ) from err

        yield from as_triple_stream(
            raw_data=meta_data,
            context=self.context,
            local_iri=self.local_iri,
            named_contexts=self.named_contexts,
        )

        # The IRI of the local page is a page.
        # FIXME: maybe this should be in octiron.py and work globally.
        yield Triple(self.local_iri, rdflib.RDF.type, OCTA.Page)

        # The page will be available on the Web under certain URL.
        if self.global_url is not None:
            yield Triple(
                self.local_iri,
                OCTA.url,
                rdflib.Literal(self.global_url),
            )



import logging
import re
import sys
from dataclasses import dataclass, field
from enum import Enum, auto
from functools import reduce
from pathlib import Path
from types import MappingProxyType
from typing import Any, Dict, Iterable, Optional, Type

import rdflib
from iolanta.context import merge
from iolanta.loaders.local_file import IsAContext, ParserNotFound
from iolanta.models import LDContext
from iolanta.namespaces import LOCAL
from iolanta.shortcuts import as_quad_stream
from ldflex import LDFlex
from octadocs.conversions import triples_to_quads
from octadocs.describe_documentation_page import describe_documentation_page
from octadocs.octiron.context_loaders import (
    context_from_json,
    context_from_yaml,
)
from octadocs.octiron.inference import (
    apply_inference_owlrl,
    apply_inference_sparql,
)
from octadocs.octiron.plugins import Loader, MarkdownLoader, YAMLLoader
from octadocs.query import _format_query_bindings
from octadocs.types import DEFAULT_NAMESPACES
from octadocs.types import LOCAL as local
from octadocs.types import Context, QueryResult
from rdflib import ConjunctiveGraph
from rdflib.plugins.sparql.processor import SPARQLResult
from typer import Typer
from urlpath import URL

if sys.version_info >= (3, 8):
    from functools import cached_property  # noqa
else:
    from backports.cached_property import cached_property  # noqa: WPS433,WPS440

logger = logging.getLogger(__name__)

CONTEXT_FORMATS = MappingProxyType({
    'context.json': context_from_json,
    'context.yaml': context_from_yaml,
})


class CacheStatus(Enum):
    """Cache condition of a file."""

    NOT_CACHED = auto()
    UP_TO_DATE = auto()
    EXPIRED = auto()


@dataclass
class Octiron:  # noqa: WPS214
    """
    Convert a lump of goo and data into a semantic graph.

    I believe that this class should be transformed into `Iolanta` class, and
    factored out of `octadocs` application into its own package.
    """

    root_directory: Path
    custom_namespaces: Dict[str, rdflib.Namespace] = field(default_factory=dict)
    last_modified_timestamp_per_file: Dict[rdflib.URIRef, float] = field(
        default_factory=dict,
        metadata={
            '__doc__': 'Time when every file was last imported into the graph.',
        },
    )
    root_context: Optional[LDContext] = None
    cli: Dict[str, Typer] = field(default_factory=dict)

    @cached_property
    def namespaces(self):
        """
        Join the provided custom namespaces with the defaults.

        Returns all namespaces registered with this Octiron instance.
        """
        namespaces = dict(DEFAULT_NAMESPACES)
        namespaces.update(self.custom_namespaces)

        return namespaces

    @cached_property
    def graph(self) -> rdflib.ConjunctiveGraph:
        """Generate and instantiate the RDFLib graph instance."""
        conjunctive_graph = ConjunctiveGraph(
            identifier=LOCAL.term('_inference'),
        )

        for short_name, namespace in self.namespaces.items():
            conjunctive_graph.bind(short_name, namespace)

        return conjunctive_graph

    def get_context_per_directory(
        self,
        directory: Path,
    ) -> Context:
        """Find context file per disk directory."""
        return reduce(
            merge,  # type: ignore
            map(
                self._get_context_file,
                reversed(list(self._find_context_files(directory))),
            ),
            self.root_context or {},
        )

    def create_file_cache_status(
        self,
        local_iri: rdflib.URIRef,
        last_modification_timestamp_on_disk: float,
    ) -> CacheStatus:
        """Determine caching status of a file."""
        cached_modification_time = self.last_modified_timestamp_per_file.get(
            local_iri,
        )

        if cached_modification_time is None:
            return CacheStatus.NOT_CACHED

        elif cached_modification_time < last_modification_timestamp_on_disk:
            return CacheStatus.EXPIRED

        return CacheStatus.UP_TO_DATE

    def clear_named_graph(self, local_iri: rdflib.URIRef) -> None:
        """Remove all triples in the specified named graph."""
        # Ugly formatting is used because of:
        #   https://github.com/RDFLib/rdflib/issues/1277
        self.graph.update(f'CLEAR GRAPH <{local_iri}>')

    def update_from_file(  # noqa: WPS210
        self,
        path: Path,
        local_iri: rdflib.URIRef,
        named_contexts: Optional[Dict[str, Any]] = None,
        global_url: Optional[str] = None,
        skip_file_if_not_importable: bool = True,
    ) -> None:
        """Update the graph from file determined by given path."""
        # Create a shorter (printable) version of the path for logging messages.
        try:
            relative_path = path.relative_to(self.root_directory)
        except ValueError:
            relative_path = path

        file_last_modification_time = path.stat().st_mtime

        cache_status = self.create_file_cache_status(
            local_iri=local_iri,
            last_modification_timestamp_on_disk=file_last_modification_time,
        )

        if cache_status == CacheStatus.UP_TO_DATE:
            logger.info('Skipping %s (cached and up to date)', relative_path)
            return

        elif cache_status == CacheStatus.EXPIRED:
            self.clear_named_graph(local_iri)

        context = self.get_context_per_directory(path.parent)
        loader_class = self.get_loader_class_for_path(path)

        if loader_class is None:
            try:
                quads = as_quad_stream(
                    url=URL(f'file://{path}'),
                    iri=local_iri,
                    default_context=self.root_context,
                    root_directory=self.root_directory,
                )
            except ParserNotFound:
                if skip_file_if_not_importable:
                    logger.info(
                        'Cannot import %s because no loader was found.',
                        path,
                    )
                    return
                else:
                    raise

            except IsAContext:
                logger.info('Cannot parse a context, sorry')
                return

        else:
            quads = self._legacy_get_quads(
                cache_status=cache_status,
                context=context,
                global_url=global_url,
                loader_class=loader_class,
                local_iri=local_iri,
                named_contexts=named_contexts,
                path=path,
                relative_path=relative_path,
            )

        self.graph.addN(quads)

        triples = describe_documentation_page(
            iri=local_iri,
            path=path,
            docs_dir=self.root_directory,
        )

        quads = triples_to_quads(
            triples=triples,
            graph=local.fs,
        )

        self.graph.addN(quads)

        # Store the file last modification time
        self.last_modified_timestamp_per_file[local_iri] = (
            file_last_modification_time
        )

    def _legacy_get_quads(
        self,
        cache_status,
        context,
        global_url,
        loader_class,
        local_iri,
        named_contexts,
        path,
        relative_path,
    ):
        """Legacy mechanism to convert a file into quad stream."""
        logger.info(
            'Importing %s via %s (%s)',
            relative_path,
            loader_class.__name__,
            'not cached before' if (
                cache_status == CacheStatus.NOT_CACHED
            ) else 'cached but expired',
        )

        loader_instance = loader_class(
            path=path,
            context=context,
            local_iri=local_iri,
            global_url=global_url,
            named_contexts=named_contexts,
        )

        triples = loader_instance.stream()
        return triples_to_quads(triples=triples, graph=local_iri)

    def apply_inference(self) -> None:  # noqa: WPS213
        """Apply inference rules after graph has been updated from files."""
        apply_inference_owlrl(graph=self.graph)

        sparql_inference_applied = apply_inference_sparql(
            inference_directory=self.root_directory.parent / 'inference',
            graph=self.graph,
        )

        self.graph.update('''
        INSERT {
            GRAPH <local:inference/page-facet/> {
                ?page octa:template ?template .
            }
        } WHERE {
            ?page iolanta:facet [
                a octa:PageFacet ;
                iolanta:supports octa: ;
                octa:template ?template
            ] .
        }
        ''')

        if sparql_inference_applied:
            # Changes made by SPARQL queries might require another pass of OWL RL
            apply_inference_owlrl(graph=self.graph)

    def get_loader_class_for_path(self, path: Path) -> Optional[Type[Loader]]:
        """Based on file path, determine the loader to use."""
        # TODO dependency inversion
        loaders = [MarkdownLoader, YAMLLoader]

        for loader in loaders:
            if re.search(loader.regex, str(path)) is not None:
                return loader

        return None

    def _find_context_files(self, directory: Path) -> Iterable[Path]:
        """
        Find all context files relevant to particular directory.

        Files are ordered from the deepest to the upmost.
        """
        for context_directory in (directory, *directory.parents):
            for filename in CONTEXT_FORMATS.keys():
                context_path = context_directory / filename

                if context_path.is_file():
                    yield context_path

            if context_directory == self.root_directory:
                return

    def _get_context_file(self, path: Path) -> Context:
        """Read and return context file by path."""
        logger.info('Loading context: %s', path)
        context_loader = CONTEXT_FORMATS[path.name]
        return context_loader(path)

    @cached_property
    def ldflex(self) -> LDFlex:
        """Construct LDFlex instance."""
        return LDFlex(self.graph)

    def query(
        self,
        query_text: str,
        **kwargs: str,
    ) -> QueryResult:
        """Run SPARQL SELECT, CONSTRUCT, or ASK query."""
        sparql_result: SPARQLResult = self.graph.query(
            query_text,
            initBindings=kwargs,
        )

        if sparql_result.askAnswer is not None:
            return sparql_result.askAnswer

        if sparql_result.graph is not None:
            graph: rdflib.Graph = sparql_result.graph
            for prefix, namespace in self.graph.namespaces():
                graph.bind(prefix, namespace)

            return graph

        return _format_query_bindings(sparql_result.bindings)



import itertools
import operator
from dataclasses import dataclass
from typing import Dict, Iterable, List

from documented import DocumentedError
from dominate.tags import li, strong, ul
from more_itertools import first
from octadocs.iolanta import render
from octadocs.octiron import Octiron
from octadocs_adr.models import ADR
from rdflib import URIRef
from rdflib.term import Node


def retrieve_properties_by_page(octiron: Octiron, iri: URIRef):
    about_this_page = octiron.query(
        '''
        SELECT ?property ?value WHERE {
            ?page ?property ?value .

            ?property a adr:ADRProperty .

            OPTIONAL {
                ?property octa:position ?explicit_position .
            }

            BIND(COALESCE(?explicit_position, 0) as ?position)
        } ORDER BY ?position ?property
        ''',
        page=iri,
    )

    groups = itertools.groupby(
        about_this_page,
        key=operator.itemgetter('property'),
    )

    return dict({
        grouper: list(map(
            operator.itemgetter('value'),
            group_items,
        ))
        for grouper, group_items in groups
    })


def render_property_values(
    property_values: List[Node],
    octiron: Octiron,
) -> str:
    rendered_values = [
        render(
            node=property_value,
            octiron=octiron,
            environment=ADR.term('sidebar-property-value'),
        )
        for property_value in property_values
    ]

    if len(rendered_values) == 1:
        return rendered_values[0]

    return ul(*map(li, rendered_values))


def render_properties_and_values(
    properties_and_values: Dict[URIRef, List[Node]],
    octiron: Octiron,
) -> Iterable[li]:
    for property_iri, property_values in properties_and_values.items():
        rendered_property = render(
            node=property_iri,
            octiron=octiron,
            environment=ADR.term('sidebar-property'),
        )

        rendered_values = render_property_values(
            property_values=property_values,
            octiron=octiron,
        )

        yield li(
            rendered_property,
            rendered_values,
            cls='md-nav__item md-nav__link',
        )


def page_sidebar(octiron: Octiron, iri: URIRef):
    """Draw list of all ADR documents."""
    properties_and_values = retrieve_properties_by_page(
        octiron=octiron,
        iri=iri,
    )

    return '\n'.join(map(str, render_properties_and_values(
        properties_and_values=properties_and_values,
        octiron=octiron,
    )))


@dataclass
class PropertyNotRenderable(DocumentedError):
    """
    Cannot render property for ADR page.

        Property IRI: {self.iri}

    Please ensure that the property has a proper `rdfs:label` assigned to it.
    """

    iri: URIRef


def sidebar_property(octiron: Octiron, iri: URIRef) -> str:
    """Render name of the property of the ADR page."""
    rows = octiron.query(
        '''
        SELECT * WHERE {
            ?property rdfs:label ?label .
            
            OPTIONAL {
                ?property octa:symbol ?symbol .
            }
        } LIMIT 1
        ''',
        property=iri,
    )

    try:
        row = first(rows)
    except ValueError as err:
        raise PropertyNotRenderable(iri=iri) from err

    label = row['label']
    if symbol := row.get('symbol'):
        label = f'{symbol} {label}'

    return strong(f'{label}: ')



from typing import Any, Dict, NamedTuple, Union

from rdflib import Literal, URIRef

# JSON-LD context

LDContext = Dict[str, Any]  # type: ignore
LDDocument = Dict[str, Any]  # type: ignore


# Named context URLs
ContextAliases = Dict[str, str]


class Triple(NamedTuple):
    """RDF triple."""

    subject: URIRef
    predicate: URIRef
    object: Union[URIRef, Literal]  # noqa: WPS125

    def as_quad(self, graph: URIRef) -> 'Quad':
        """Add graph to this triple and hence get a quad."""
        return Quad(
            subject=self.subject,
            predicate=self.predicate,
            object=self.object,
            graph=graph,
        )


class Quad(NamedTuple):
    """Triple assigned to a named graph."""

    subject: URIRef
    predicate: URIRef
    object: Union[URIRef, Literal]  # noqa: WPS125
    graph: URIRef

    def as_triple(self):
        """Convert this to triple."""
        return Triple(self.subject, self.predicate, self.object)

    def __repr__(self):
        return (
            f'(<{self.subject}> <{self.predicate}> <{self.object}> @ '
            f'{self.graph})'
        )



import itertools
import logging
import operator
from dataclasses import dataclass
from functools import cached_property
from typing import Dict, List, Optional

from dominate import tags
from mkdocs.structure.pages import Page
from more_itertools import first
from octadocs.iolanta import render
from octadocs.octiron import Octiron
from rdflib import URIRef
from rdflib.term import Literal, Node

logger = logging.getLogger(__name__)


def app_by_property(octiron: Octiron) -> Dict[URIRef, URIRef]:
    """Find apps connected to properties."""
    pairs = octiron.query(
        '''
        SELECT ?property ?app WHERE {
            ?property iolanta:facet ?facet .
            ?app iolanta:supports adr:sidebar .
        }
        '''
    )
    return {
        row['property']: row['facet']
        for row in pairs
    }


def default_property_facet(
    octiron: Octiron,
    property_iri: URIRef,
    property_values: List[Node],
) -> Optional[str]:
    """Default facet to render a property with its values."""
    label = first(
        map(
            operator.itemgetter('label'),
            octiron.query(
                '''
                SELECT ?label WHERE {
                    ?property rdfs:label ?label .
                } ORDER BY ?label LIMIT 1
                ''',
                property=property_iri,
            ),
        ),
        None,
    )

    if label is None:
        return

    if len(property_values) > 1:
        raise Exception('Too many values')

    property_value = property_values[0]

    if isinstance(property_value, Literal):
        return f'<strong>{label}</strong>: {property_value}'

    rendered_value = render(
        octiron=octiron,
        node=property_value,
    )
    return f'<strong>{label}</strong>: {rendered_value}'


@dataclass
class DecisionContext:
    """Context for the Decision template."""

    page: Page
    octiron: Octiron

    @cached_property
    def iri(self) -> URIRef:
        """Retrieve the IRI of current page."""
        return self.page.iri

    @cached_property
    def status(self) -> str:
        status_choices = self.octiron.query('''
            SELECT ?label ?symbol WHERE {
                ?page adr:status [
                    rdfs:label ?label ;
                    octa:symbol ?symbol
                ] .
            }
        ''', page=self.iri)

        if not status_choices:
            return ''

        elif len(status_choices) < 2:
            status, = status_choices
            label = status['label']
            symbol = status['symbol']
            return f'{symbol} {label}'

        raise ValueError(
            f'This page has too many status values: {status_choices}',
        )

    @cached_property
    def author(self):
        authors = self.octiron.query('''
            SELECT ?name ?url WHERE {
                ?page adr:author [
                    schema:name ?name ;
                    schema:url ?url
                ] .
            }
        ''', page=self.iri)

        if not authors:
            return ''

        if len(authors) < 2:
            author, = authors

            dom = tags.li(
                tags.a(
                    tags.strong('Author: '),
                    author['name'],

                    cls='md-nav__link',
                    href=author['url'],
                    target='_blank',
                ),
                cls='md-nav__item',
            )

            return str(dom)

        raise ValueError('page has too many authors!')

    @cached_property
    def date(self) -> str:
        date_choice = self.octiron.query('''
            SELECT ?date WHERE {
                ?page adr:date ?date .
            }
        ''', page=self.iri).first

        if not date_choice:
            return ''

        return date_choice['date'].value

    def describe_this_page(self) -> Dict[URIRef, List[Node]]:
        """List all properties of current page that can be rendered."""
        about_this_page = self.octiron.query(
            '''
            SELECT ?property ?value WHERE {
                ?page ?property ?value .

                ?property a adr:ADRProperty .

                OPTIONAL {
                    ?property octa:position ?explicit_position .
                }

                BIND(COALESCE(?explicit_position, 0) as ?position)
            } ORDER BY ?position ?property
            ''',
            page=self.page.iri,
        )

        groups = itertools.groupby(
            about_this_page,
            key=operator.itemgetter('property'),
        )

        return dict({
            grouper: list(map(
                operator.itemgetter('value'),
                group_items,
            ))
            for grouper, group_items in groups
        })

    def page_properties(self):
        """Render page properties block."""
        properties_and_apps = app_by_property(self.octiron)
        properties_and_values = self.describe_this_page()

        for property_iri, property_values in properties_and_values.items():
            if attached_app_iri := properties_and_apps.get(property_iri):
                raise ValueError(f'Attached app: {attached_app_iri}')

            rendered_facet = default_property_facet(
                octiron=self.octiron,
                property_iri=property_iri,
                property_values=property_values,
            )

            yield f'<li class="md-nav__item md-nav__link">{rendered_facet}</li>'

    @property
    def render_page_properties(self):
        return '\n'.join(filter(None, self.page_properties()))



from pathlib import Path
from typing import Any, Dict

from mkdocs.plugins import BasePlugin
from mkdocs.structure.pages import Page
from octadocs.octiron.context_loaders import context_from_yaml
from octadocs.plugin import cached_octiron
from octadocs_table.models import TABLE
from rdflib import URIRef


class TablePlugin(BasePlugin):
    """Render an HTML table from data presented in the graph."""

    @property
    def templates_path(self) -> Path:
        """Templates associated with the plugin."""
        return Path(__file__).parent / 'templates'

    def context_url(self) -> str:
        path = Path(__file__).parent / 'yaml/ctx.yaml'
        return f'file://{path}'

    def load_context(self):
        """Load YAML-LD context."""
        return context_from_yaml(Path(__file__).parent / 'yaml/ctx.yaml')

    def on_config(self, config, **kwargs):
        """Adjust configuration."""
        # Make plugin's templates available to MkDocs
        config['theme'].dirs.append(str(self.templates_path))

        named_context = self.load_context()

        try:
            contexts = config['extra']['named_contexts']
        except KeyError:
            contexts = config['extra']['named_contexts'] = {}

        contexts['table'] = named_context

        docs_dir = Path(config['docs_dir'])

        self.octiron = cached_octiron(
            docs_dir=docs_dir,
        )

        # Prefix
        self.octiron.graph.bind('table', TABLE)

        # Load the triples
        self.octiron.update_from_file(
            path=Path(__file__).parent / 'yaml/octadocs-table.yaml',
            local_iri=URIRef(TABLE),
            global_url='/octadocs-table.yaml',
            named_contexts=config['extra']['named_contexts'],
        )

    def on_page_context(
        self,
        context: Dict[str, Any],
        page: Page,
        **kwargs,
    ):
        """Make custom functions available to the template."""
        context.update({
            'table': TABLE,
        })

        return context



from typing import Iterator

import yaml
from octadocs.octiron.plugins import Loader
from octadocs.octiron.yaml_extensions import SafeLoader, as_triple_stream
from octadocs.types import OCTA, Triple
from rdflib import RDF


class YAMLLoader(Loader):
    """Load semantic data from Markdown front matter."""

    regex = r'\.ya*ml$'

    def stream(self) -> Iterator[Triple]:
        """Return stream of triples."""
        if self.path.stem == 'context':
            return

        with open(self.path, 'r') as yaml_file:
            raw_data = yaml.load(yaml_file, Loader=SafeLoader)

        if raw_data is None:
            return

        yield from as_triple_stream(
            raw_data=raw_data,
            context=self.context,
            local_iri=self.local_iri,
            named_contexts=self.named_contexts,
        )

        # The IRI of the local page is a page.
        # FIXME: maybe this should be in octiron.py and work globally.
        yield Triple(self.local_iri, RDF.type, OCTA.Page)



from abc import ABC
from dataclasses import dataclass, field
from typing import Iterable, Optional, TextIO

from iolanta.models import ContextAliases, LDContext, LDDocument, Quad
from rdflib import URIRef
from urlpath import URL


@dataclass(frozen=True)
class Parser(ABC):
    """
    Parser reads data from a file-like object and interprets them.

    For interpretation, it is also supplied with a context.
    """

    blank_node_prefix: str = ''

    def as_jsonld_document(self, raw_data: TextIO) -> LDDocument:
        """Generate a JSON-LD document."""
        raise NotImplementedError()

    def as_quad_stream(
        self,
        raw_data: TextIO,
        iri: Optional[URIRef],
        context: LDContext,
    ) -> Iterable[Quad]:
        raise NotImplementedError()




from copy import deepcopy

from deepmerge import always_merger
from octadocs.types import Context


def merge(first: Context, second: Context) -> Context:
    """
    Merge two contexts into one.

    Second context can override the first.
    """
    return deepcopy(always_merger.merge(
        base=first,
        nxt=second,
    ))



import re

from octadocs.types import LOCAL


def iri_to_url(iri: str) -> str:
    """Convert Zet IRI into clickable URL."""
    iri = iri.replace(str(LOCAL), '')

    if iri.endswith('index.md'):
        return re.sub(
            r'/index\.md$',
            '/',
            iri,
        )

    else:
        return re.sub(
            r'\.md$',
            '',
            iri,
        )



from dominate.tags import span
from iolanta.facet import Facet
from more_itertools import first


class GithubLicense(Facet):
    """Render license information."""

    def render(self):
        """Render."""
        rows = self.query(
            '''
            SELECT * WHERE {
                $license
                    gh:spdx_id ?id ;
                    gh:name ?name .
            }
            ''',
            license=self.iri,
        )

        row = first(rows)

        title = row['name']
        text = str(row['id'])

        if text == 'NOASSERTION':
            text = title

        return span(text, title=title)



from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Dict

from iolanta.context import merge
from iolanta.models import LDContext
from octadocs.pyld_document_loaders.base import DocumentLoader
from octadocs.pyld_document_loaders.models import JsonLDDocument


@dataclass
class MergeLoader(DocumentLoader):
    """
    Merge context retrieved from another source with a pre-generated context.

    For example, load context from a remote URL and then merge with context
    of current directory.
    """

    base_context: LDContext
    loader: DocumentLoader

    def __call__(self, url: str, options: Dict[str, Any]) -> JsonLDDocument:
        child_document = self.loader(url=url, options=options)
        base_context = deepcopy(self.base_context)

        if '@import' in base_context:
            base_context.pop('@import')

        base_context = {
            '@context': base_context,
        }

        child_document['document'] = merge(
            first=base_context,
            second=child_document['document'],
        )

        # raise ValueError(child_document)

        return child_document



import operator
from functools import cached_property
from pathlib import Path
from typing import Dict, Optional, Union

from iolanta.loaders.base import term_for_python_class
from iolanta.models import LDContext, Triple
from mkdocs.plugins import BasePlugin
from octadocs.octiron import Octiron
from octadocs.plugin import cached_octiron
from rdflib import Namespace, URIRef
from rdflib.term import Node
from typer import Typer


class OctadocsMixin(BasePlugin):
    """MkDocs plugin that aims to extend Octadocs functionality."""

    octiron: Octiron
    namespaces: Optional[Dict[str, Namespace]] = None
    plugin_data_dir: Path

    def typer(self) -> Optional[Typer]:
        """Return a CLI command for octadocs app."""

    @cached_property
    def templates_path(self) -> Optional[Path]:
        """Templates associated with the plugin."""
        path = Path(__file__).parent / 'templates'
        if path.exists():
            return path

    def named_contexts(self) -> Dict[str, LDContext]:
        """Named contexts."""
        return {}

    def vocabularies(self) -> Dict[URIRef, Path]:
        """Pieces of structured data to load into graph."""
        return {}

    def on_config(self, config, **kwargs):
        """Adjust system configuration to suit this plugin."""
        # Make plugin's templates available to MkDocs
        if self.templates_path:
            config['theme'].dirs.append(str(self.templates_path))

        docs_dir = Path(config['docs_dir'])
        self.octiron = cached_octiron(
            docs_dir=docs_dir,
        )
        self.bind_namespaces()

        # Publish named contexts to make them reusable for the whole system
        config['extra']['named_contexts'].update(self.named_contexts())

        # Load vocabularies and data into graph
        for url, path in self.vocabularies().items():
            self.octiron.update_from_file(
                path=path,
                local_iri=url,
                global_url=url,
                named_contexts=config['extra']['named_contexts'],
                skip_file_if_not_importable=False,
            )

    def on_files(self, *args, **kwargs):
        """Extract metadata from files and compose the site graph."""
        self.inference()

    def inference(self):
        """Apply inference, if any."""

    def update(self, sparql_query: str):
        """Apply the given SPARQL INSERT query."""
        self.octiron.graph.update(sparql_query)

    def query(self, sparql_query: str, **kwargs: Union[str, int, Node]):
        """Query and return results."""
        return self.octiron.query(sparql_query, **kwargs)

    def insert(self, *triples: Triple):
        """Insert triples into graph."""
        graph = term_for_python_class(self.__class__)
        quads = map(
            operator.methodcaller('as_quad', graph),
            triples,
        )
        self.octiron.graph.addN(quads)

    def bind_namespaces(self):
        if not self.namespaces:
            return

        for prefix, namespace in self.namespaces.items():
            self.octiron.graph.bind(prefix, namespace)



from datetime import date
from typing import Union

from classes import typeclass


@typeclass
def render_literal_value(literal_value) -> str:
    """Render a literal value nicely for printing."""


@render_literal_value.instance(None)
def _render_none(literal_value: None) -> str:
    return '∅ None'


@render_literal_value.instance(bool)
def _render_bool(literal_value: bool) -> str:
    icon = '✅' if literal_value else '❌'
    return f'{icon} {literal_value}'


@render_literal_value.instance(int)
def _render_int(literal_value: Union[int, float]) -> str:
    return f'🔢 {literal_value}'


@render_literal_value.instance(str)
def _render_str(literal_value: str) -> str:
    return f'🔡 {literal_value}'


@render_literal_value.instance(date)
def _render_date(literal_value: date) -> str:
    return f'📅 {literal_value}'


@render_literal_value.instance(object)
def _render_default(literal_value: object) -> str:
    return f'❓ {literal_value}'



from functools import partial
from typing import Any, Dict, Optional, Union

import rdflib
from mkdocs_macros.plugin import MacrosPlugin
from octadocs.conversions import iri_by_page, src_path_to_iri
from octadocs.environment import iri_to_url
from octadocs.iolanta import render
from octadocs.octiron import Octiron
from octadocs.query import query
from octadocs.types import LOCAL, OCTA
from rdflib.plugins.sparql.processor import SPARQLResult


def sparql(
    instance: rdflib.ConjunctiveGraph,
    query: str,
    **kwargs: str,
) -> SPARQLResult:
    bindings = {
        argument_name: argument_value
        for argument_name, argument_value in kwargs.items()
    }

    return instance.query(query, initBindings=bindings)


def _render_as_row(row: Dict[rdflib.Variable, Any]) -> str:  # type: ignore
    """Render row of a Markdown table."""
    formatted_row = ' | '.join(row.values())
    return f'| {formatted_row} |'


def table(query_result: SPARQLResult) -> str:
    """Render as a Markdown table."""
    headers = ' | '.join(str(cell) for cell in query_result.vars)

    rows = '\n'.join(
        _render_as_row(row)
        for row in query_result.bindings
    )

    separators = '| ' + (' --- |' * len(query_result.vars))  # noqa: WPS336

    return f'''
---
| {headers} |
{separators}
{rows}
'''


def url(
    resource: rdflib.URIRef,
    graph: rdflib.ConjunctiveGraph
) -> Optional[str]:
    """Convert a URIRef to a clickable URL."""
    bindings = graph.query(
        'SELECT ?url WHERE { ?resource octa:subjectOf/octa:url ?url . } ',
        initBindings={
            'resource': resource,
        }
    ).bindings

    if not bindings:
        return None

    return '/' + bindings[0][rdflib.Variable('url')].value


def label(
    resource: rdflib.URIRef,
    graph: rdflib.ConjunctiveGraph
) -> Optional[str]:
    """Convert a URIRef to a clickable URL."""
    bindings = graph.query(
        'SELECT ?label WHERE { ?resource rdfs:label ?label . } ',
        initBindings={
            'resource': resource,
        }
    ).bindings

    if not bindings:
        return None

    return bindings[0][rdflib.Variable('label')].value


def link(reference: Union[str, rdflib.URIRef], octiron: Octiron) -> str:
    """Render link to the object."""
    if isinstance(reference, str):
        if ':' in reference:
            reference = rdflib.URIRef(reference)
        else:
            reference = LOCAL.term(reference)

    return render(
        node=reference,
        octiron=octiron,
        environments=[OCTA.link],
    )


def define_env(env: MacrosPlugin) -> MacrosPlugin:  # noqa: WPS213
    """Create mkdocs-macros Jinja environment."""
    env.filter(sparql)
    env.filter(table)

    octiron = env.variables.octiron
    env.variables.update(octiron.namespaces)
    env.variables['LOCAL'] = LOCAL
    env.variables['local'] = LOCAL

    env.macro(
        partial(
            link,
            octiron=octiron,
        ),
        name='link',
    )

    env.macro(
        partial(
            render,
            octiron=octiron,
        ),
        name='render',
    )

    env.macro(
        partial(
            query,
            instance=env.variables.graph,
        ),
        name='query',
    )

    env.macro(
        partial(
            url,
            graph=env.variables.graph,
        ),
        name='url',
    )

    env.filter(
        partial(
            url,
            graph=env.variables.graph,
        ), name='url',
    )

    env.filter(
        partial(
            label,
            graph=env.variables.graph,
        ),
        name='label',
    )

    env.macro(iri_to_url)
    env.macro(src_path_to_iri)

    env.filter(iri_to_url)

    env.filter(iri_by_page)
    env.macro(iri_by_page)

    # Update context with namespaces
    env.variables.update(octiron.namespaces)

    env.variables['URIRef'] = rdflib.URIRef

    return env



import operator
from typing import List, Optional, TypedDict

from dominate.tags import li, table, tbody, td, th, thead, tr, ul
from dominate.util import raw
from iolanta.facet import Facet
from iolanta.namespaces import IOLANTA
from octadocs.iolanta import render
from octadocs_table.models import TABLE
from rdflib.term import Node


class Row(TypedDict):
    """Raw response."""

    prov_value: Node
    comment: Optional[str]
    count: int


class ColumnLegend(Facet):
    """
    Render a legend for a column.

    FIXME: This facet could have been implemented in a more generic way.

    ```yaml
    $id: RDFLibSupportLegend
    $type: ColumnLegend
    column: supports-rdflib
    ```

    Via OWL rules, this construct would generate a new `Table` node in the graph
    which would be dynamically rendered.

    Unfortunately, current feature set of `octadocs-table` does not permit that.
    In particular,

    - we do not have `table:self` to make system render the node itself,
    - `table:class` property of a table is required, cannot draw a table
      without it,
    - YAML description of a table cannot describe aggregation functions.

    We will be implementing those points over time and perhaps this widget will
    be thereupon refactored.
    """

    legend_query = '''
    SELECT ?prov_value ?comment (COUNT(?prov_value) AS ?count) WHERE {
        ?instance $column ?node .

        {
            FILTER isLiteral(?node) .
        } UNION {
            FILTER (isIRI(?node) || isBlank(?node)) .
            ?node prov:value+ ?prov_value .

            OPTIONAL {
                {
                    ?prov_value rdfs:comment ?comment .
                } UNION {
                    ?node rdfs:comment ?comment .
                }
            }
        }
    }
    GROUP BY ?prov_value ?comment
    ORDER BY DESC(?count)
    '''

    def render_comment(self, row: Row):
        """Render description column."""
        comment = row.get('comment', '') or ''

        if comment:
            yield raw(comment)

        if row['prov_value'] is None:
            return

        see_also_links = list(
            map(
                operator.itemgetter('link'),
                self.query(
                    '''
                    SELECT ?link WHERE {
                        $node rdfs:seeAlso ?link .

                        OPTIONAL {
                            $node octa:position ?position .
                        }
                    }
                    ORDER BY ?position
                    ''',
                    node=row['prov_value'],
                ),
            ),
        )

        if see_also_links:
            yield ul(
                li(
                    render(
                        node=link,
                        octiron=self.octiron,
                    ),
                )
                for link in see_also_links
            )

    def render(self):
        """Print unique values for a column."""
        rows: List[Row] = self.query(
            query_text=self.legend_query,
            column=self.iri,
        )

        table_rows = [
            tr(
                td(
                    raw(
                        render(
                            node=row['prov_value'],
                            octiron=self.octiron,
                            environments=[
                                TABLE.td,
                                IOLANTA.html,
                            ],
                        ) if row['prov_value'] is not None else '',
                    ),
                ),
                td(*self.render_comment(row)),
                td(row['count']),
            )
            for row in rows
        ]

        return table(
            thead(
                tr(
                    # These are hard coded, and I cannot change that. See the
                    # docstring for details.
                    th('Value'),
                    th('Description'),
                    th('Count'),
                ),
            ),
            tbody(*table_rows),
        )



"""
Iolanta facet management.

This module contains a few functions which later will be refactored into
Iolanta - the generic metaverse/cyberspace browser.
"""
import operator
import pydoc
from dataclasses import dataclass
from typing import Callable, Iterable, List, Optional, Union

from iolanta.facet import Facet
from iolanta.facets.base import FacetSearchAttempt
from iolanta.facets.by_environment import FindFacetByEnvironment
from iolanta.facets.by_instance import FindFacetByInstance
from iolanta.facets.by_literal_datatype import FindFacetByLiteralDatatype
from iolanta.facets.by_type import FindFacetByType
from ldflex import LDFlex
from octadocs.iolanta.errors import FacetError, FacetNotCallable, FacetNotFound
from octadocs.octiron import Octiron
from rdflib import RDFS
from rdflib.term import Literal, Node, URIRef

HTML = URIRef('https://html.spec.whatwg.org/')


def resolve_facet(iri: URIRef) -> Callable[[Octiron, Node], str]:
    """Resolve a path to a Python object to that object."""
    url = str(iri)

    if not url.startswith('python://'):
        raise Exception(
            'Octadocs only supports facets which are importable Python '
            'callables. The URLs of such facets must start with `python://`, '
            'which {url} does not comply to.'.format(
                url=url,
            )
        )

    # It is impossible to use `urlpath` for this operation because it (or,
    # rather, one of upper classes from `urllib` that `urlpath` depends upon)
    # will lowercase the URL when parsing it - which means, irreversibly. We
    # have to resort to plain string manipulation.
    import_path = url.replace('python://', '').strip('/')

    facet = pydoc.locate(import_path)

    if not callable(facet):
        raise FacetNotCallable(
            path=import_path,
            facet=facet,
        )

    return facet


def render(
    node: Union[str, Node],
    octiron: Octiron,
    environments: Optional[List[URIRef]] = None,
) -> str:
    """Find an Iolanta facet for a node and render it."""
    if not environments:
        environments = [HTML]

    facet_search_attempt = Render(
        ldflex=octiron.ldflex,
    ).search_for_facet(
        node=node,
        environments=environments,
    )

    facet = resolve_facet(
        iri=facet_search_attempt.facet,
    )

    facet = facet(
        octiron=octiron,
        iri=node,
        environment=facet_search_attempt.environment,
    )

    if isinstance(facet, Facet):
        try:
            return facet.render()

        except (FacetError, FacetNotFound):   # noqa: WPS329
            # Prevent nested `FacetError` exceptions.
            raise

        except Exception as err:
            raise FacetError(
                node=node,
                facet_iri=facet_search_attempt.facet,
                facet_search_attempt=facet_search_attempt,
                error=err,
            ) from err

    return facet


@dataclass
class Render:
    """Facet renderer."""

    ldflex: LDFlex

    def search_for_facet(
        self,
        node: Node,
        environments: List[URIRef],
    ) -> FacetSearchAttempt:
        """Find facet IRI for given node for all environments given."""
        facet_search_attempts = list(
            self.attempt_search_for_facet(
                node=node,
                environments=environments,
            ),
        )

        *failures, success = facet_search_attempts

        if success:
            return success

        raise FacetNotFound(
            node=node,
            environments=environments,
            facet_search_attempts=facet_search_attempts,
        )

    def find_facet_iri(
        self,
        node: Node,
        environments: List[URIRef],
    ) -> URIRef:
        """Find facet IRI for given node for all environments given."""
        return self.search_for_facet(
            node=node,
            environments=environments,
        ).facet

    def attempt_search_for_facet(
        self,
        node: Node,
        environments: List[URIRef],
    ) -> Iterable[FacetSearchAttempt]:
        """
        Stream of attempts to find a facet.

        If this function yields a sequence of N elements, first (N - 1) of those
        are guaranteed to be failures. The remaining one element can be both
        a failure (if nothing was found) or a success (if a facet was found).
        """
        for environment in environments:
            facet_searches = self.find_facet_iri_per_environment(
                node=node,
                environment=environment,
            )

            for facet_search in facet_searches:
                yield facet_search

                if facet_search:
                    return

    def find_facet_iri_per_environment(
        self,
        node: Node,
        environment: URIRef,
    ):
        """Find facet IRI for given node in given env."""
        if isinstance(node, Literal):
            yield from self.find_facet_iri_for_literal(
                literal=node,
                environment=environment,
            )

        yield FindFacetByInstance(
            ldflex=self.ldflex,
            node=node,
            environment=environment,
        )

        instance_types = self.find_instance_types(node)
        for instance_type in instance_types:
            yield FindFacetByType(
                ldflex=self.ldflex,
                node=node,
                environment=environment,
                instance_type=instance_type,
            )

        yield FindFacetByEnvironment(
            ldflex=self.ldflex,
            node=node,
            environment=environment,
        )

    def __call__(
        self,
        node: Node,
        environment: URIRef,
    ) -> str:
        ...

    def find_facet_iri_for_literal(self, literal: Literal, environment: URIRef):
        """Find facet IRI for a literal."""
        if literal.datatype is not None:
            yield FindFacetByLiteralDatatype(
                ldflex=self.ldflex,
                node=literal,
                environment=environment,
            )

        yield FindFacetByType(
            ldflex=self.ldflex,
            node=literal,
            environment=environment,
            instance_type=RDFS.Literal,
        )

        yield FindFacetByEnvironment(
            ldflex=self.ldflex,
            node=literal,
            environment=environment,
        )

    def find_instance_types(self, node: URIRef) -> List[URIRef]:
        """Find types for particular node."""
        rows = self.ldflex.query(
            query_text='''
            SELECT ?type WHERE {
                $node rdf:type ?type .
            }
            ''',
            node=node,
        )

        return list(
            map(
                operator.itemgetter('type'),
                rows,
            ),
        )
